@@ -1,4 +1,4 @@
-# ruff: noqa: S603, PTH118
+# ruff: noqa: S603
 from __future__ import annotations
 
 import builtins
@@ -23,6 +23,10 @@ if _git_bin is None:
 GIT_BIN_PATH = _git_bin
 
 
+def restart() -> None:
+    os.execvpe(sys.executable, sys.orig_argv, os.environ)  # noqa: S606
+
+
 def git_check_ignore(rc_path: StrPath) -> None:
     rc_path = os.fspath(rc_path)
     subprocess.check_output([GIT_BIN_PATH, "-C", git_root(), "check-ignore", rc_path])
@@ -35,8 +39,12 @@ def git_root() -> str:
     ).strip()
 
 
-def project_rc_path(basename: str = PYTHONRC_BASENAME) -> str:
-    return os.path.relpath(os.path.join(git_root(), basename))
+def project_rc_path(basename: str = PYTHONRC_BASENAME) -> str | None:
+    try:
+        root = os.path.join(git_root(), basename)
+    except subprocess.CalledProcessError:
+        return None
+    return os.path.relpath(root, start=git_root())
 
 
 @contextmanager
@@ -90,8 +98,10 @@ def init_rc_script(
     rc_path: StrPath,
     global_ns: dict[str, object],
 ) -> None:
-    execute_rc_script(rc_path, global_ns)
-    global_ns["reload"] = reload_function(rc_path, global_ns)
+    try:
+        execute_rc_script(rc_path, global_ns)
+    finally:
+        global_ns["reload"] = reload_function(rc_path, global_ns)
 
 
 def clean_module_cache() -> None:
@@ -128,9 +138,36 @@ class DisplayHookPatcher:
         if not hasattr(self, "original_hook"):
             msg = "DisplayHookPatcher is not started. Call start() first."
             raise RuntimeError(msg)
-        if self.active:
+        if self.active and not sys.exception():
             if obj is not None:
                 self.printer(obj)
                 set_last_value(obj)
         else:
             self.original_hook(obj)
+
+
+def generate_stub_for_pythonstartup(out_dir: StrPath | None = None) -> bool:
+    link = out_dir is None
+    out_dir = os.path.dirname(os.fspath(out_dir) if out_dir else __file__)
+
+    startup_file = os.environ["PYTHONSTARTUP"]
+
+    try:
+        from mypy import stubgen  # noqa: PLC0415
+    except ImportError:
+        return False
+
+    stubgen.main(
+        [
+            startup_file,
+            "--output",
+            out_dir,
+            "--ignore-errors",
+            "--no-import",
+            "--parse-only",
+            "--include-private",
+        ]
+    )
+    if link:
+        os.symlink(startup_file, os.path.join(out_dir, os.path.basename(startup_file)))
+    return True
